@@ -2098,7 +2098,7 @@ getchar_common(typval_T *argvars, typval_T *rettv)
 
     // redraw the screen after getchar()
     if (p_ch == 0)
-	update_screen(CLEAR);
+	update_screen(UPD_CLEAR);
 
     set_vim_var_nr(VV_MOUSE_WIN, 0);
     set_vim_var_nr(VV_MOUSE_WINID, 0);
