@@ -96,11 +96,11 @@
 #ifdef SIGHASARG
 # ifdef SIGHAS3ARGS
 #  define SIGPROTOARG	(int, int, struct sigcontext *)
-#  define SIGDEFARG(s)	(s, sig2, scont) int s, sig2; struct sigcontext *scont;
+#  define SIGDEFARG(s)	(int s, int sig2, struct sigcontext *scont)
 #  define SIGDUMMYARG	0, 0, (struct sigcontext *)0
 # else
 #  define SIGPROTOARG	(int)
-#  define SIGDEFARG(s)	(s) int s UNUSED;
+#  define SIGDEFARG(s)	(int s UNUSED)
 #  define SIGDUMMYARG	0
 # endif
 #else
@@ -128,10 +128,7 @@
 # endif
 #endif
 
-// on some systems time.h should not be included together with sys/time.h
-#if !defined(HAVE_SYS_TIME_H) || defined(TIME_WITH_SYS_TIME)
-# include <time.h>
-#endif
+#include <time.h>
 #ifdef HAVE_SYS_TIME_H
 # include <sys/time.h>
 #endif
